#!/usr/bin/python3
#    Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This prometheus-proxy is intended to abstract the prometheus metrics
# exported from the reference provider driver load balancing engines (haproxy
# and lvs) such that all of the provider drivers can expose a consistent set
# of metrics. It also aligns the terms to be consistent with Octavia
# terminology.

from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler
import os
import signal
import socketserver
import sys
import threading
import time
import traceback
import urllib.request

import psutil

from octavia.amphorae.backends.utils import network_namespace
from octavia.common import constants as consts

METRICS_URL = "http://127.0.0.1:9101/metrics"
PRINT_REJECTED = False
EXIT_EVENT = threading.Event()

# A dictionary of prometheus metrics mappings.
# Key: The metric string to match
# Value: A tuple of replacement data.
#    tuple[0]: The string to replace the key with.
#    tuple[1]: If not None, the replacement HELP line for the metric.
#    If tuple[1] is None, the key will be replaced in the HELP string.
#    tuple[2]: If not None, includes a dictionary of additional substitutions.
#    tuple[2] substitutions happen prior to the key replacement in tuple[0].
METRIC_MAP = {
    # Load balancer metrics
    "haproxy_process_pool_allocated_bytes ":
        ("octavia_memory_pool_allocated_bytes ",
         "# HELP octavia_memory_pool_allocated_bytes Total amount of memory "
         "allocated in the memory pools (in bytes).\n", None),
    "haproxy_process_pool_used_bytes ":
        ("octavia_memory_pool_used_bytes ",
         "# HELP octavia_memory_pool_used_bytes Total amount of memory used "
         "in the memory pools (in bytes).\n", None),
    "haproxy_process_pool_failures_total ":
        ("octavia_memory_pool_failures_total ",
         "# HELP octavia_memory_pool_failures_total Total number of failed "
         "memory pool allocations.\n", None),
    "haproxy_process_max_connections ":
        ("octavia_loadbalancer_max_connections ", None, None),
    "haproxy_process_current_connections ":
        ("octavia_loadbalancer_current_connections ", None, None),
    # TODO(johnsom) consider adding in UDP
    "haproxy_process_connections_total ":
        ("octavia_loadbalancer_connections_total ", None, None),
    # TODO(johnsom) consider adding in UDP (and update help string)
    "haproxy_process_requests_total ":
        ("octavia_loadbalancer_requests_total ", None, None),
    "haproxy_process_max_ssl_connections ":
        ("octavia_loadbalancer_max_ssl_connections ", None, None),
    "haproxy_process_current_ssl_connections ":
        ("octavia_loadbalancer_current_ssl_connections ",
         "# HELP octavia_loadbalancer_current_ssl_connections Number of "
         "active SSL connections.\n", None),
    "haproxy_process_ssl_connections_total ":
        ("octavia_loadbalancer_ssl_connections_total ", None, None),
    "haproxy_process_current_connection_rate ":
        ("octavia_loadbalancer_current_connection_rate ", None, None),
    "haproxy_process_limit_connection_rate ":
        ("octavia_loadbalancer_limit_connection_rate ", None, None),
    "haproxy_process_max_connection_rate ":
        ("octavia_loadbalancer_max_connection_rate ", None, None),
    "haproxy_process_current_session_rate ":
        ("octavia_loadbalancer_current_session_rate ", None, None),
    "haproxy_process_limit_session_rate ":
        ("octavia_loadbalancer_limit_session_rate ", None, None),
    "haproxy_process_max_session_rate ":
        ("octavia_loadbalancer_max_session_rate ", None, None),
    "haproxy_process_current_ssl_rate ":
        ("octavia_loadbalancer_current_ssl_rate ", None, None),
    "haproxy_process_limit_ssl_rate ":
        ("octavia_loadbalancer_limit_ssl_rate ", None, None),
    "haproxy_process_max_ssl_rate ":
        ("octavia_loadbalancer_max_ssl_rate ", None, None),
    "haproxy_process_current_frontend_ssl_key_rate ":
        ("octavia_loadbalancer_current_frontend_ssl_key_rate ", None, None),
    "haproxy_process_max_frontend_ssl_key_rate ":
        ("octavia_loadbalancer_max_frontend_ssl_key_rate ", None, None),
    "haproxy_process_frontent_ssl_reuse ":
        ("octavia_loadbalancer_frontent_ssl_reuse ", None, None),
    "haproxy_process_current_backend_ssl_key_rate ":
        ("octavia_loadbalancer_current_backend_ssl_key_rate ", None, None),
    "haproxy_process_max_backend_ssl_key_rate ":
        ("octavia_loadbalancer_max_backend_ssl_key_rate ", None, None),
    "haproxy_process_ssl_cache_lookups_total ":
        ("octavia_loadbalancer_ssl_cache_lookups_total ", None, None),
    "haproxy_process_ssl_cache_misses_total ":
        ("octavia_loadbalancer_ssl_cache_misses_total ", None, None),
    "haproxy_process_http_comp_bytes_in_total ":
        ("octavia_loadbalancer_http_comp_bytes_in_total ", None, None),
    "haproxy_process_http_comp_bytes_out_total ":
        ("octavia_loadbalancer_http_comp_bytes_out_total ", None, None),
    "haproxy_process_limit_http_comp ":
        ("octavia_loadbalancer_limit_http_comp ", None, None),
    "haproxy_process_listeners ":
        ("octavia_loadbalancer_listeners ", None, None),
    "haproxy_process_dropped_logs_total ":
        ("octavia_loadbalancer_dropped_logs_total ", None, None),

    # Listener metrics
    "haproxy_frontend_status ":
        ("octavia_listener_status ",
         "# HELP octavia_listener_status Current status of the listener. "
         "0=OFFLINE, 1=ONLINE, 2=DEGRADED.\n", None),
    "haproxy_frontend_status{":
        ("octavia_listener_status{", None, {"proxy=": "listener="}),
    "haproxy_frontend_current_sessions ":
        ("octavia_listener_current_sessions ", None, None),
    "haproxy_frontend_current_sessions{":
        ("octavia_listener_current_sessions{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_max_sessions ":
        ("octavia_listener_max_sessions ", None, None),
    "haproxy_frontend_max_sessions{":
        ("octavia_listener_max_sessions{", None, {"proxy=": "listener="}),
    "haproxy_frontend_limit_sessions ":
        ("octavia_listener_limit_sessions ", None, None),
    "haproxy_frontend_limit_sessions{":
        ("octavia_listener_limit_sessions{", None, {"proxy=": "listener="}),
    "haproxy_frontend_sessions_total ":
        ("octavia_listener_sessions_total ", None, None),
    "haproxy_frontend_sessions_total{":
        ("octavia_listener_sessions_total{", None, {"proxy=": "listener="}),
    "haproxy_frontend_limit_session_rate ":
        ("octavia_listener_limit_session_rate ", None, None),
    "haproxy_frontend_limit_session_rate{":
        ("octavia_listener_limit_session_rate{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_max_session_rate ":
        ("octavia_listener_max_session_rate ", None, None),
    "haproxy_frontend_max_session_rate{":
        ("octavia_listener_max_session_rate{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_connections_rate_max ":
        ("octavia_listener_connections_rate_max ", None, None),
    "haproxy_frontend_connections_rate_max{":
        ("octavia_listener_connections_rate_max{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_connections_total ":
        ("octavia_listener_connections_total ", None, None),
    "haproxy_frontend_connections_total{":
        ("octavia_listener_connections_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_bytes_in_total ":
        ("octavia_listener_bytes_in_total ", None, None),
    "haproxy_frontend_bytes_in_total{":
        ("octavia_listener_bytes_in_total{", None, {"proxy=": "listener="}),
    "haproxy_frontend_bytes_out_total ":
        ("octavia_listener_bytes_out_total ", None, None),
    "haproxy_frontend_bytes_out_total{":
        ("octavia_listener_bytes_out_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_requests_denied_total ":
        ("octavia_listener_requests_denied_total ", None, None),
    "haproxy_frontend_requests_denied_total{":
        ("octavia_listener_requests_denied_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_responses_denied_total ":
        ("octavia_listener_responses_denied_total ", None, None),
    "haproxy_frontend_responses_denied_total{":
        ("octavia_listener_responses_denied_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_request_errors_total ":
        ("octavia_listener_request_errors_total ", None, None),
    "haproxy_frontend_request_errors_total{":
        ("octavia_listener_request_errors_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_denied_connections_total ":
        ("octavia_listener_denied_connections_total ",
         "# HELP octavia_listener_denied_connections_total Total number of "
         "requests denied by connection rules.\n", None),
    "haproxy_frontend_denied_connections_total{":
        ("octavia_listener_denied_connections_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_denied_sessions_total ":
        ("octavia_listener_denied_sessions_total ",
         "# HELP octavia_listener_denied_sessions_total Total number of "
         "requests denied by session rules.\n", None),
    "haproxy_frontend_denied_sessions_total{":
        ("octavia_listener_denied_sessions_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_failed_header_rewriting_total ":
        ("octavia_listener_failed_header_rewriting_total ",
         "# HELP octavia_listener_failed_header_rewriting_total Total number "
         "of failed header rewriting rules.\n", None),
    "haproxy_frontend_failed_header_rewriting_total{":
        ("octavia_listener_failed_header_rewriting_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_requests_rate_max ":
        ("octavia_listener_http_requests_rate_max ", None, None),
    "haproxy_frontend_http_requests_rate_max{":
        ("octavia_listener_http_requests_rate_max{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_requests_total ":
        ("octavia_listener_http_requests_total ", None, None),
    "haproxy_frontend_http_requests_total{":
        ("octavia_listener_http_requests_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_responses_total ":
        ("octavia_listener_http_responses_total ", None, None),
    "haproxy_frontend_http_responses_total{":
        ("octavia_listener_http_responses_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_intercepted_requests_total ":
        ("octavia_listener_intercepted_requests_total ", None, None),
    "haproxy_frontend_intercepted_requests_total{":
        ("octavia_listener_intercepted_requests_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_cache_lookups_total ":
        ("octavia_listener_http_cache_lookups_total ", None, None),
    "haproxy_frontend_http_cache_lookups_total{":
        ("octavia_listener_http_cache_lookups_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_cache_hits_total ":
        ("octavia_listener_http_cache_hits_total ", None, None),
    "haproxy_frontend_http_cache_hits_total{":
        ("octavia_listener_http_cache_hits_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_comp_bytes_in_total ":
        ("octavia_listener_http_comp_bytes_in_total ", None, None),
    "haproxy_frontend_http_comp_bytes_in_total{":
        ("octavia_listener_http_comp_bytes_in_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_comp_bytes_out_total ":
        ("octavia_listener_http_comp_bytes_out_total ", None, None),
    "haproxy_frontend_http_comp_bytes_out_total{":
        ("octavia_listener_http_comp_bytes_out_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_comp_bytes_bypassed_total ":
        ("octavia_listener_http_comp_bytes_bypassed_total ", None, None),
    "haproxy_frontend_http_comp_bytes_bypassed_total{":
        ("octavia_listener_http_comp_bytes_bypassed_total{", None,
         {"proxy=": "listener="}),
    "haproxy_frontend_http_comp_responses_total ":
        ("octavia_listener_http_comp_responses_total ", None, None),
    "haproxy_frontend_http_comp_responses_total{":
        ("octavia_listener_http_comp_responses_total{", None,
         {"proxy=": "listener="}),

    # Pool Metrics
    "haproxy_backend_status ":
        ("octavia_pool_status ",
         "# HELP octavia_pool_status Current status of the pool. 0=OFFLINE, "
         "1=ONLINE.\n", None),
    "haproxy_backend_status{":
        ("octavia_pool_status{", None, {"proxy=": "pool="}),
    "haproxy_backend_current_sessions ":
        ("octavia_pool_current_sessions ", None, None),
    "haproxy_backend_current_sessions{":
        ("octavia_pool_current_sessions{", None, {"proxy=": "pool="}),
    "haproxy_backend_max_sessions ":
        ("octavia_pool_max_sessions ", None, None),
    "haproxy_backend_max_sessions{":
        ("octavia_pool_max_sessions{", None, {"proxy=": "pool="}),
    "haproxy_backend_limit_sessions ":
        ("octavia_pool_limit_sessions ", None, None),
    "haproxy_backend_limit_sessions{":
        ("octavia_pool_limit_sessions{", None, {"proxy=": "pool="}),
    "haproxy_backend_sessions_total ":
        ("octavia_pool_sessions_total ", None, None),
    "haproxy_backend_sessions_total{":
        ("octavia_pool_sessions_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_max_session_rate ":
        ("octavia_pool_max_session_rate ", None, None),
    "haproxy_backend_max_session_rate{":
        ("octavia_pool_max_session_rate{", None, {"proxy=": "pool="}),
    "haproxy_backend_last_session_seconds ":
        ("octavia_pool_last_session_seconds ",
         "# HELP octavia_pool_last_session_seconds Number of seconds since "
         "last session assigned to a member.\n", None, None),
    "haproxy_backend_last_session_seconds{":
        ("octavia_pool_last_session_seconds{", None, {"proxy=": "pool="}),
    "haproxy_backend_current_queue ":
        ("octavia_pool_current_queue ", None, None),
    "haproxy_backend_current_queue{":
        ("octavia_pool_current_queue{", None, {"proxy=": "pool="}),
    "haproxy_backend_max_queue ":
        ("octavia_pool_max_queue ", None, None),
    "haproxy_backend_max_queue{":
        ("octavia_pool_max_queue{", None, {"proxy=": "pool="}),
    "haproxy_backend_connection_attempts_total ":
        ("octavia_pool_connection_attempts_total ", None, None),
    "haproxy_backend_connection_attempts_total{":
        ("octavia_pool_connection_attempts_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_connection_reuses_total ":
        ("octavia_pool_connection_reuses_total ", None, None),
    "haproxy_backend_connection_reuses_total{":
        ("octavia_pool_connection_reuses_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_bytes_in_total ":
        ("octavia_pool_bytes_in_total ", None, None),
    "haproxy_backend_bytes_in_total{":
        ("octavia_pool_bytes_in_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_bytes_out_total ":
        ("octavia_pool_bytes_out_total ", None, None),
    "haproxy_backend_bytes_out_total{":
        ("octavia_pool_bytes_out_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_queue_time_average_seconds ":
        ("octavia_pool_queue_time_average_seconds ", None, None),
    "haproxy_backend_queue_time_average_seconds{":
        ("octavia_pool_queue_time_average_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_connect_time_average_seconds ":
        ("octavia_pool_connect_time_average_seconds ", None, None),
    "haproxy_backend_connect_time_average_seconds{":
        ("octavia_pool_connect_time_average_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_response_time_average_seconds ":
        ("octavia_pool_response_time_average_seconds ", None, None),
    "haproxy_backend_response_time_average_seconds{":
        ("octavia_pool_response_time_average_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_total_time_average_seconds ":
        ("octavia_pool_total_time_average_seconds ", None, None),
    "haproxy_backend_total_time_average_seconds{":
        ("octavia_pool_total_time_average_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_max_queue_time_seconds ":
        ("octavia_pool_max_queue_time_seconds ", None, None),
    "haproxy_backend_max_queue_time_seconds{":
        ("octavia_pool_max_queue_time_seconds{", None, {"proxy=": "pool="}),
    "haproxy_backend_max_connect_time_seconds ":
        ("octavia_pool_max_connect_time_seconds ", None, None),
    "haproxy_backend_max_connect_time_seconds{":
        ("octavia_pool_max_connect_time_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_max_response_time_seconds ":
        ("octavia_pool_max_response_time_seconds ",
         "# HELP octavia_pool_max_response_time_seconds Maximum observed "
         "time spent waiting for a member response.\n", None),
    "haproxy_backend_max_response_time_seconds{":
        ("octavia_pool_max_response_time_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_max_total_time_seconds ":
        ("octavia_pool_max_total_time_seconds ", None, None),
    "haproxy_backend_max_total_time_seconds{":
        ("octavia_pool_max_total_time_seconds{", None, {"proxy=": "pool="}),
    "haproxy_backend_requests_denied_total ":
        ("octavia_pool_requests_denied_total ", None, None),
    "haproxy_backend_requests_denied_total{":
        ("octavia_pool_requests_denied_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_responses_denied_total ":
        ("octavia_pool_responses_denied_total ", None, None),
    "haproxy_backend_responses_denied_total{":
        ("octavia_pool_responses_denied_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_connection_errors_total ":
        ("octavia_pool_connection_errors_total ", None, None),
    "haproxy_backend_connection_errors_total{":
        ("octavia_pool_connection_errors_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_response_errors_total ":
        ("octavia_pool_response_errors_total ", None, None),
    "haproxy_backend_response_errors_total{":
        ("octavia_pool_response_errors_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_retry_warnings_total ":
        ("octavia_pool_retry_warnings_total ", None, None),
    "haproxy_backend_retry_warnings_total{":
        ("octavia_pool_retry_warnings_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_redispatch_warnings_total ":
        ("octavia_pool_redispatch_warnings_total ", None, None),
    "haproxy_backend_redispatch_warnings_total{":
        ("octavia_pool_redispatch_warnings_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_failed_header_rewriting_total ":
        ("octavia_pool_failed_header_rewriting_total ", None, None),
    "haproxy_backend_failed_header_rewriting_total{":
        ("octavia_pool_failed_header_rewriting_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_client_aborts_total ":
        ("octavia_pool_client_aborts_total ", None, None),
    "haproxy_backend_client_aborts_total{":
        ("octavia_pool_client_aborts_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_server_aborts_total ":
        ("octavia_pool_member_aborts_total ",
         "# HELP octavia_pool_server_aborts_total Total number of data "
         "transfers aborted by the server.\n", None),
    "haproxy_backend_server_aborts_total{":
        ("octavia_pool_member_aborts_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_active_servers ":
        ("octavia_pool_active_members ",
         "# HELP octavia_pool_active_members Current number of active "
         "members.\n", None),
    "haproxy_backend_active_servers{":
        ("octavia_pool_active_members{", None, {"proxy=": "pool="}),
    "haproxy_backend_backup_servers ":
        ("octavia_pool_backup_members ",
         "# HELP octavia_pool_backup_members Current number of backup "
         "members.\n", None),
    "haproxy_backend_backup_servers{":
        ("octavia_pool_backup_members{", None, {"proxy=": "pool="}),
    "haproxy_backend_check_up_down_total ":
        ("octavia_pool_check_up_down_total ", None, None),
    "haproxy_backend_check_up_down_total{":
        ("octavia_pool_check_up_down_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_check_last_change_seconds ":
        ("octavia_pool_check_last_change_seconds ", None, None),
    "haproxy_backend_check_last_change_seconds{":
        ("octavia_pool_check_last_change_seconds{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_downtime_seconds_total ":
        ("octavia_pool_downtime_seconds_total ",
         "# HELP octavia_pool_downtime_seconds_total Total downtime "
         "(in seconds) for the pool.\n", None),
    "haproxy_backend_downtime_seconds_total{":
        ("octavia_pool_downtime_seconds_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_loadbalanced_total ":
        ("octavia_pool_loadbalanced_total ",
         "# HELP octavia_pool_loadbalanced_total Total number of times a "
         "pool was selected, either for new sessions, or when "
         "redispatching.\n", None),
    "haproxy_backend_loadbalanced_total{":
        ("octavia_pool_loadbalanced_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_http_requests_total ":
        ("octavia_pool_http_requests_total ", None, None),
    "haproxy_backend_http_requests_total{":
        ("octavia_pool_http_requests_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_http_responses_total ":
        ("octavia_pool_http_responses_total ", None, None),
    "haproxy_backend_http_responses_total{":
        ("octavia_pool_http_responses_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_http_cache_lookups_total ":
        ("octavia_pool_http_cache_lookups_total ", None, None),
    "haproxy_backend_http_cache_lookups_total{":
        ("octavia_pool_http_cache_lookups_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_http_cache_hits_total ":
        ("octavia_pool_http_cache_hits_total ", None, None),
    "haproxy_backend_http_cache_hits_total{":
        ("octavia_pool_http_cache_hits_total{", None, {"proxy=": "pool="}),
    "haproxy_backend_http_comp_bytes_in_total ":
        ("octavia_pool_http_comp_bytes_in_total ", None, None),
    "haproxy_backend_http_comp_bytes_in_total{":
        ("octavia_pool_http_comp_bytes_in_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_http_comp_bytes_out_total ":
        ("octavia_pool_http_comp_bytes_out_total ", None, None),
    "haproxy_backend_http_comp_bytes_out_total{":
        ("octavia_pool_http_comp_bytes_out_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_http_comp_bytes_bypassed_total ":
        ("octavia_pool_http_comp_bytes_bypassed_total ", None, None),
    "haproxy_backend_http_comp_bytes_bypassed_total{":
        ("octavia_pool_http_comp_bytes_bypassed_total{", None,
         {"proxy=": "pool="}),
    "haproxy_backend_http_comp_responses_total ":
        ("octavia_pool_http_comp_responses_total ", None, None),
    "haproxy_backend_http_comp_responses_total{":
        ("octavia_pool_http_comp_responses_total{", None,
         {"proxy=": "pool="}),

    # Member Metrics
    "haproxy_server_status ":
        ("octavia_member_status ",
         "# HELP octavia_member_status Current status of the member. "
         "0=ERROR, 1=ONLINE, 2=OFFLINE, 3=DRAIN.\n", None),
    "haproxy_server_status{":
        ("octavia_member_status{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_current_sessions ":
        ("octavia_member_current_sessions ", None, None),
    "haproxy_server_current_sessions{":
        ("octavia_member_current_sessions{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_sessions ":
        ("octavia_member_max_sessions ", None, None),
    "haproxy_server_max_sessions{":
        ("octavia_member_max_sessions{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_limit_sessions ":
        ("octavia_member_limit_sessions ", None, None),
    "haproxy_server_limit_sessions{":
        ("octavia_member_limit_sessions{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_sessions_total ":
        ("octavia_member_sessions_total ", None, None),
    "haproxy_server_sessions_total{":
        ("octavia_member_sessions_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_session_rate ":
        ("octavia_member_max_session_rate ", None, None),
    "haproxy_server_max_session_rate{":
        ("octavia_member_max_session_rate{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_last_session_seconds ":
        ("octavia_member_last_session_seconds ",
         "# HELP octavia_member_last_session_seconds Number of seconds since "
         "last session assigned to the member.\n", None),
    "haproxy_server_last_session_seconds{":
        ("octavia_member_last_session_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_current_queue ":
        ("octavia_member_current_queue ", None, None),
    "haproxy_server_current_queue{":
        ("octavia_member_current_queue{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_queue ":
        ("octavia_member_max_queue ", None, None),
    "haproxy_server_max_queue{":
        ("octavia_member_max_queue{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_queue_limit ":
        ("octavia_member_queue_limit ",
         "# HELP octavia_member_queue_limit Configured maxqueue for the "
         "member (0 meaning no limit).\n", None),
    "haproxy_server_queue_limit{":
        ("octavia_member_queue_limit{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_bytes_in_total ":
        ("octavia_member_bytes_in_total ", None, None),
    "haproxy_server_bytes_in_total{":
        ("octavia_member_bytes_in_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_bytes_out_total ":
        ("octavia_member_bytes_out_total ", None, None),
    "haproxy_server_bytes_out_total{":
        ("octavia_member_bytes_out_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_queue_time_average_seconds ":
        ("octavia_member_queue_time_average_seconds ", None, None),
    "haproxy_server_queue_time_average_seconds{":
        ("octavia_member_queue_time_average_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_connect_time_average_seconds ":
        ("octavia_member_connect_time_average_seconds ", None, None),
    "haproxy_server_connect_time_average_seconds{":
        ("octavia_member_connect_time_average_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_response_time_average_seconds ":
        ("octavia_member_response_time_average_seconds ", None, None),
    "haproxy_server_response_time_average_seconds{":
        ("octavia_member_response_time_average_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_total_time_average_seconds ":
        ("octavia_member_total_time_average_seconds ", None, None),
    "haproxy_server_total_time_average_seconds{":
        ("octavia_member_total_time_average_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_queue_time_seconds ":
        ("octavia_member_max_queue_time_seconds ", None, None),
    "haproxy_server_max_queue_time_seconds{":
        ("octavia_member_max_queue_time_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_connect_time_seconds ":
        ("octavia_member_max_connect_time_seconds ", None, None),
    "haproxy_server_max_connect_time_seconds{":
        ("octavia_member_max_connect_time_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_response_time_seconds ":
        ("octavia_member_max_response_time_seconds ",
         "# HELP octavia_member_max_response_time_seconds Maximum observed "
         "time spent waiting for a member response.\n", None),
    "haproxy_server_max_response_time_seconds{":
        ("octavia_member_max_response_time_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_max_total_time_seconds ":
        ("octavia_member_max_total_time_seconds ", None, None),
    "haproxy_server_max_total_time_seconds{":
        ("octavia_member_max_total_time_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_connection_attempts_total ":
        ("octavia_member_connection_attempts_total ", None, None),
    "haproxy_server_connection_attempts_total{":
        ("octavia_member_connection_attempts_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_connection_reuses_total ":
        ("octavia_member_connection_reuses_total ", None, None),
    "haproxy_server_connection_reuses_total{":
        ("octavia_member_connection_reuses_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_responses_denied_total ":
        ("octavia_member_responses_denied_total ", None, None),
    "haproxy_server_responses_denied_total{":
        ("octavia_member_responses_denied_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_connection_errors_total ":
        ("octavia_member_connection_errors_total ", None, None),
    "haproxy_server_connection_errors_total{":
        ("octavia_member_connection_errors_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_response_errors_total ":
        ("octavia_member_response_errors_total ", None, None),
    "haproxy_server_response_errors_total{":
        ("octavia_member_response_errors_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_retry_warnings_total ":
        ("octavia_member_retry_warnings_total ", None, None),
    "haproxy_server_retry_warnings_total{":
        ("octavia_member_retry_warnings_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_redispatch_warnings_total ":
        ("octavia_member_redispatch_warnings_total ", None, None),
    "haproxy_server_redispatch_warnings_total{":
        ("octavia_member_redispatch_warnings_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_failed_header_rewriting_total ":
        ("octavia_member_failed_header_rewriting_total ", None, None),
    "haproxy_server_failed_header_rewriting_total{":
        ("octavia_member_failed_header_rewriting_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_client_aborts_total ":
        ("octavia_member_client_aborts_total ", None, None),
    "haproxy_server_client_aborts_total{":
        ("octavia_member_client_aborts_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_server_aborts_total ":
        ("octavia_member_server_aborts_total ", None, None),
    "haproxy_server_server_aborts_total{":
        ("octavia_member_server_aborts_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_weight ":
        ("octavia_member_weight ",
         "# HELP octavia_member_weight Member weight.\n", None),
    "haproxy_server_weight{":
        ("octavia_member_weight{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_check_failures_total ":
        ("octavia_member_check_failures_total ",
         "# HELP octavia_member_check_failures_total Total number of failed "
         "check (Only counts checks failed when the member is up).\n", None),
    "haproxy_server_check_failures_total{":
        ("octavia_member_check_failures_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_check_up_down_total ":
        ("octavia_member_check_up_down_total ", None, None),
    "haproxy_server_check_up_down_total{":
        ("octavia_member_check_up_down_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_downtime_seconds_total ":
        ("octavia_member_downtime_seconds_total ",
         "# HELP octavia_member_downtime_seconds_total Total downtime (in "
         "seconds) for the member.\n", None),
    "haproxy_server_downtime_seconds_total{":
        ("octavia_member_downtime_seconds_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_check_last_change_seconds ":
        ("octavia_member_check_last_change_seconds ", None, None),
    "haproxy_server_check_last_change_seconds{":
        ("octavia_member_check_last_change_seconds{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_current_throttle ":
        ("octavia_member_current_throttle ",
         "# HELP octavia_member_current_throttle Current throttle percentage "
         "for the member, when slowstart is active, or no value if not in "
         "slowstart.\n", None),
    "haproxy_server_current_throttle{":
        ("octavia_member_current_throttle{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_loadbalanced_total ":
        ("octavia_member_loadbalanced_total ",
         "# HELP octavia_member_loadbalanced_total Total number of times a "
         "member was selected, either for new sessions, or when "
         "redispatching.\n", None),
    "haproxy_server_loadbalanced_total{":
        ("octavia_member_loadbalanced_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_http_responses_total ":
        ("octavia_member_http_responses_total ", None, None),
    "haproxy_server_http_responses_total{":
        ("octavia_member_http_responses_total{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_server_idle_connections_current ":
        ("octavia_member_idle_connections_current ", None, None),
    "haproxy_server_server_idle_connections_current{":
        ("octavia_member_idle_connections_current{", None,
         {"proxy=": "pool=", "server=": "member="}),
    "haproxy_server_server_idle_connections_limit ":
        ("octavia_member_idle_connections_limit ", None, None),
    "haproxy_server_server_idle_connections_limit{":
        ("octavia_member_idle_connections_limit{", None,
         {"proxy=": "pool=", "server=": "member="}),
}
METRIC_KEYS = METRIC_MAP.keys()


class PrometheusProxy(SimpleHTTPRequestHandler):

    protocol_version = 'HTTP/1.1'

    # No need to log every request through the proxy
    def log_request(self, *args, **kwargs):
        pass

    def _add_cpu_utilization(self, metrics_buffer):
        cpu_pcnt = (psutil.getloadavg()[0] / os.cpu_count()) * 100
        metrics_buffer += ("# HELP octavia_loadbalancer_cpu Load balancer "
                           "CPU utilization (percentage).\n")
        metrics_buffer += "# TYPE octavia_loadbalancer_cpu gauge\n"
        cpu_metric_string = f"octavia_loadbalancer_cpu {cpu_pcnt:.1f}\n"
        metrics_buffer += cpu_metric_string
        return metrics_buffer

    def _add_memory_utilization(self, metrics_buffer):
        mem_pcnt = psutil.virtual_memory()[2]
        metrics_buffer += ("# HELP octavia_loadbalancer_memory Load balancer "
                           "memory utilization (percentage).\n")
        metrics_buffer += "# TYPE octavia_loadbalancer_memory gauge\n"
        mem_metric_string = f"octavia_loadbalancer_memory {mem_pcnt:.1f}\n"
        metrics_buffer += mem_metric_string
        return metrics_buffer

    def do_GET(self):
        metrics_buffer = ""

        metrics_buffer = self._add_cpu_utilization(metrics_buffer)
        metrics_buffer = self._add_memory_utilization(metrics_buffer)

        try:
            with urllib.request.urlopen(METRICS_URL) as source:  # nosec
                lines = source.readlines()
                for line in lines:
                    line = line.decode("utf-8")
                    # Don't report metrics for the internal prometheus
                    # proxy loop. The user facing listener will still be
                    # reported.
                    if "prometheus-exporter" in line:
                        continue
                    match = next((x for x in METRIC_KEYS if x in line), False)
                    if match:
                        map_tuple = METRIC_MAP[match]
                        if map_tuple[1] and "HELP" in line:
                            metrics_buffer += map_tuple[1]
                        else:
                            if map_tuple[2] and not line.startswith("#"):
                                for key in map_tuple[2].keys():
                                    line = line.replace(key,
                                                        map_tuple[2][key])
                            metrics_buffer += line.replace(match,
                                                           map_tuple[0])
                    elif PRINT_REJECTED:
                        print("REJECTED: %s" % line)
        except Exception as e:
            print(str(e), flush=True)
            traceback.print_tb(e.__traceback__)
            self.send_response(502)
            self.send_header("connection", "close")
            self.end_headers()
            return

        self.send_response(200)
        self.send_header("cache-control", "no-cache")
        self.send_header("content-type", "text/plain; version=0.0.4")
        self.send_header("connection", "close")
        self.end_headers()
        self.wfile.write(metrics_buffer.encode("utf-8"))


class SignalHandler:

    def __init__(self):
        signal.signal(signal.SIGINT, self.shutdown)
        signal.signal(signal.SIGTERM, self.shutdown)

    def shutdown(self, *args):
        EXIT_EVENT.set()


def shutdown_thread(http):
    EXIT_EVENT.wait()
    http.shutdown()


# TODO(johnsom) Remove and switch to ThreadingHTTPServer once python3.7 is
# the minimum version supported.
class ThreadedHTTPServer(socketserver.ThreadingMixIn, HTTPServer):
    daemon_threads = True


def main():
    global PRINT_REJECTED
    try:
        if sys.argv[1] == "--rejected":
            PRINT_REJECTED = True
    except Exception:
        pass

    SignalHandler()

    while not EXIT_EVENT.is_set():
        # The amphora-haproxy network namespace may not be present, so handle
        # it gracefully.
        try:
            with network_namespace.NetworkNamespace(consts.AMPHORA_NAMESPACE):
                httpd = ThreadedHTTPServer(('127.0.0.1', 9102),
                                           PrometheusProxy)
                shutdownthread = threading.Thread(target=shutdown_thread,
                                                  args=(httpd,))
                shutdownthread.start()

                # TODO(johnsom) Uncomment this when we move to
                #               ThreadingHTTPServer
                # httpd.daemon_threads = True
                print("Now serving on port 9102")
                httpd.serve_forever()
        except Exception:
            time.sleep(1)


if __name__ == "__main__":
    main()



# Copyright 2018 Rackspace, US Inc.
# Copyright 2019 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import errno
import os
import socketserver
import threading

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from octavia.api.drivers.driver_agent import driver_get
from octavia.api.drivers.driver_agent import driver_updater


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _recv(recv_socket):
    size_str = b''
    char = recv_socket.recv(1)
    while char != b'\n':
        size_str += char
        char = recv_socket.recv(1)
    payload_size = int(size_str)
    mv_buffer = memoryview(bytearray(payload_size))
    next_offset = 0
    while payload_size - next_offset > 0:
        recv_size = recv_socket.recv_into(mv_buffer[next_offset:],
                                          payload_size - next_offset)
        next_offset += recv_size
    return jsonutils.loads(mv_buffer.tobytes())


class StatusRequestHandler(socketserver.BaseRequestHandler):

    def handle(self):
        # Get the update data
        status = _recv(self.request)

        # Process the update
        updater = driver_updater.DriverUpdater()
        response = updater.update_loadbalancer_status(status)

        # Send the response
        json_data = jsonutils.dump_as_bytes(response)
        len_str = '{}\n'.format(len(json_data)).encode('utf-8')
        self.request.send(len_str)
        self.request.sendall(json_data)


class StatsRequestHandler(socketserver.BaseRequestHandler):

    def handle(self):
        # Get the update data
        stats = _recv(self.request)

        # Process the update
        updater = driver_updater.DriverUpdater()
        response = updater.update_listener_statistics(stats)

        # Send the response
        json_data = jsonutils.dump_as_bytes(response)
        len_str = '{}\n'.format(len(json_data)).encode('utf-8')
        self.request.send(len_str)
        self.request.sendall(json_data)


class GetRequestHandler(socketserver.BaseRequestHandler):

    def handle(self):
        # Get the data request
        get_data = _recv(self.request)

        # Process the get
        response = driver_get.process_get(get_data)

        # Send the response
        json_data = jsonutils.dump_as_bytes(response)
        len_str = '{}\n'.format(len(json_data)).encode('utf-8')
        self.request.send(len_str)
        self.request.sendall(json_data)


class ForkingUDSServer(socketserver.ForkingMixIn,
                       socketserver.UnixStreamServer):
    pass


def _cleanup_socket_file(filename):
    # Remove the socket file if it already exists
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def status_listener(exit_event):
    _cleanup_socket_file(CONF.driver_agent.status_socket_path)

    server = ForkingUDSServer(CONF.driver_agent.status_socket_path,
                              StatusRequestHandler)

    server.timeout = CONF.driver_agent.status_request_timeout
    server.max_children = CONF.driver_agent.status_max_processes

    while not exit_event.is_set():
        server.handle_request()

    LOG.info('Waiting for driver status listener to shutdown...')
    # Can't shut ourselves down as we would deadlock, spawn a thread
    threading.Thread(target=server.shutdown).start()
    LOG.info('Driver status listener shutdown finished.')
    server.server_close()
    _cleanup_socket_file(CONF.driver_agent.status_socket_path)


def stats_listener(exit_event):
    _cleanup_socket_file(CONF.driver_agent.stats_socket_path)

    server = ForkingUDSServer(CONF.driver_agent.stats_socket_path,
                              StatsRequestHandler)

    server.timeout = CONF.driver_agent.stats_request_timeout
    server.max_children = CONF.driver_agent.stats_max_processes

    while not exit_event.is_set():
        server.handle_request()

    LOG.info('Waiting for driver statistics listener to shutdown...')
    # Can't shut ourselves down as we would deadlock, spawn a thread
    threading.Thread(target=server.shutdown).start()
    LOG.info('Driver statistics listener shutdown finished.')
    server.server_close()
    _cleanup_socket_file(CONF.driver_agent.stats_socket_path)


def get_listener(exit_event):
    _cleanup_socket_file(CONF.driver_agent.get_socket_path)

    server = ForkingUDSServer(CONF.driver_agent.get_socket_path,
                              GetRequestHandler)

    server.timeout = CONF.driver_agent.get_request_timeout
    server.max_children = CONF.driver_agent.get_max_processes

    while not exit_event.is_set():
        server.handle_request()

    LOG.info('Waiting for driver get listener to shutdown...')
    # Can't shut ourselves down as we would deadlock, spawn a thread
    threading.Thread(target=server.shutdown).start()
    LOG.info('Driver get listener shutdown finished.')
    server.server_close()
    _cleanup_socket_file(CONF.driver_agent.get_socket_path)
    LOG.info("UDS server was closed and socket was cleaned up.")



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from cryptography import fernet
from oslo_config import cfg
from oslo_log import log as logging
from stevedore import driver as stevedore_driver
from taskflow import task
from taskflow.types import failure

from octavia.amphorae.backends.agent import agent_jinja_cfg
from octavia.amphorae.driver_exceptions import exceptions as driver_except
from octavia.common import constants
from octavia.common import utils
from octavia.controller.worker import task_utils as task_utilities
from octavia.db import api as db_apis
from octavia.db import repositories as repo

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class BaseAmphoraTask(task.Task):
    """Base task to load drivers common to the tasks."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.amphora_driver = stevedore_driver.DriverManager(
            namespace='octavia.amphora.drivers',
            name=CONF.controller_worker.amphora_driver,
            invoke_on_load=True
        ).driver
        self.amphora_repo = repo.AmphoraRepository()
        self.listener_repo = repo.ListenerRepository()
        self.loadbalancer_repo = repo.LoadBalancerRepository()
        self.task_utils = task_utilities.TaskUtils()


class AmpListenersUpdate(BaseAmphoraTask):
    """Task to update the listeners on one amphora."""

    def execute(self, loadbalancer, amphora, timeout_dict=None):
        # Note, we don't want this to cause a revert as it may be used
        # in a failover flow with both amps failing. Skip it and let
        # health manager fix it.
        try:
            # Make sure we have a fresh load balancer object
            loadbalancer = self.loadbalancer_repo.get(db_apis.get_session(),
                                                      id=loadbalancer.id)
            self.amphora_driver.update_amphora_listeners(
                loadbalancer, amphora, timeout_dict)
        except Exception as e:
            LOG.error('Failed to update listeners on amphora %s. Skipping '
                      'this amphora as it is failing to update due to: %s',
                      amphora.id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora.id,
                                     status=constants.ERROR)


class AmphoraIndexListenerUpdate(BaseAmphoraTask):
    """Task to update the listeners on one amphora."""

    def execute(self, loadbalancer, amphora_index, amphorae,
                timeout_dict=None):
        # Note, we don't want this to cause a revert as it may be used
        # in a failover flow with both amps failing. Skip it and let
        # health manager fix it.
        try:
            # Make sure we have a fresh load balancer object
            loadbalancer = self.loadbalancer_repo.get(db_apis.get_session(),
                                                      id=loadbalancer.id)
            self.amphora_driver.update_amphora_listeners(
                loadbalancer, amphorae[amphora_index], timeout_dict)
        except Exception as e:
            amphora_id = amphorae[amphora_index].id
            LOG.error('Failed to update listeners on amphora %s. Skipping '
                      'this amphora as it is failing to update due to: %s',
                      amphora_id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                     status=constants.ERROR)


class ListenersUpdate(BaseAmphoraTask):
    """Task to update amphora with all specified listeners' configurations."""

    def execute(self, loadbalancer):
        """Execute updates per listener for an amphora."""
        self.amphora_driver.update(loadbalancer)

    def revert(self, loadbalancer, *args, **kwargs):
        """Handle failed listeners updates."""

        LOG.warning("Reverting listeners updates.")

        for listener in loadbalancer.listeners:
            self.task_utils.mark_listener_prov_status_error(listener.id)


class ListenersStart(BaseAmphoraTask):
    """Task to start all listeners on the vip."""

    def execute(self, loadbalancer, amphora=None):
        """Execute listener start routines for listeners on an amphora."""
        if loadbalancer.listeners:
            self.amphora_driver.start(loadbalancer, amphora)
            LOG.debug("Started the listeners on the vip")

    def revert(self, loadbalancer, *args, **kwargs):
        """Handle failed listeners starts."""

        LOG.warning("Reverting listeners starts.")
        for listener in loadbalancer.listeners:
            self.task_utils.mark_listener_prov_status_error(listener.id)


class AmphoraIndexListenersReload(BaseAmphoraTask):
    """Task to reload all listeners on an amphora."""

    def execute(self, loadbalancer, amphora_index, amphorae,
                timeout_dict=None):
        """Execute listener reload routines for listeners on an amphora."""
        if loadbalancer.listeners:
            try:
                self.amphora_driver.reload(
                    loadbalancer, amphorae[amphora_index], timeout_dict)
            except Exception as e:
                amphora_id = amphorae[amphora_index].id
                LOG.warning('Failed to reload listeners on amphora %s. '
                            'Skipping this amphora as it is failing to '
                            'reload due to: %s', amphora_id, str(e))
                self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                         status=constants.ERROR)


class ListenerDelete(BaseAmphoraTask):
    """Task to delete the listener on the vip."""

    def execute(self, listener):
        """Execute listener delete routines for an amphora."""
        # TODO(rm_work): This is only relevant because of UDP listeners now.
        self.amphora_driver.delete(listener)
        LOG.debug("Deleted the listener on the vip")

    def revert(self, listener, *args, **kwargs):
        """Handle a failed listener delete."""

        LOG.warning("Reverting listener delete.")

        self.task_utils.mark_listener_prov_status_error(listener.id)


class AmphoraGetInfo(BaseAmphoraTask):
    """Task to get information on an amphora."""

    def execute(self, amphora):
        """Execute get_info routine for an amphora."""
        self.amphora_driver.get_info(amphora)


class AmphoraGetDiagnostics(BaseAmphoraTask):
    """Task to get diagnostics on the amphora and the loadbalancers."""

    def execute(self, amphora):
        """Execute get_diagnostic routine for an amphora."""
        self.amphora_driver.get_diagnostics(amphora)


class AmphoraFinalize(BaseAmphoraTask):
    """Task to finalize the amphora before any listeners are configured."""

    def execute(self, amphora):
        """Execute finalize_amphora routine."""
        self.amphora_driver.finalize_amphora(amphora)
        LOG.debug("Finalized the amphora.")

    def revert(self, result, amphora, *args, **kwargs):
        """Handle a failed amphora finalize."""
        if isinstance(result, failure.Failure):
            return
        LOG.warning("Reverting amphora finalize.")
        self.task_utils.mark_amphora_status_error(amphora.id)


class AmphoraPostNetworkPlug(BaseAmphoraTask):
    """Task to notify the amphora post network plug."""

    def execute(self, amphora, ports):
        """Execute post_network_plug routine."""
        for port in ports:
            self.amphora_driver.post_network_plug(amphora, port)
            LOG.debug("post_network_plug called on compute instance "
                      "%(compute_id)s for port %(port_id)s",
                      {"compute_id": amphora.compute_id, "port_id": port.id})

    def revert(self, result, amphora, *args, **kwargs):
        """Handle a failed post network plug."""
        if isinstance(result, failure.Failure):
            return
        LOG.warning("Reverting post network plug.")
        self.task_utils.mark_amphora_status_error(amphora.id)


class AmphoraePostNetworkPlug(BaseAmphoraTask):
    """Task to notify the amphorae post network plug."""

    def execute(self, loadbalancer, added_ports):
        """Execute post_network_plug routine."""
        amp_post_plug = AmphoraPostNetworkPlug()
        # We need to make sure we have the fresh list of amphora
        amphorae = self.amphora_repo.get_all(
            db_apis.get_session(), load_balancer_id=loadbalancer.id,
            status=constants.AMPHORA_ALLOCATED)[0]
        for amphora in amphorae:
            if amphora.id in added_ports:
                amp_post_plug.execute(amphora, added_ports[amphora.id])

    def revert(self, result, loadbalancer, added_ports, *args, **kwargs):
        """Handle a failed post network plug."""
        if isinstance(result, failure.Failure):
            return
        LOG.warning("Reverting post network plug.")

        amphorae = self.amphora_repo.get_all(
            db_apis.get_session(), load_balancer_id=loadbalancer.id,
            status=constants.AMPHORA_ALLOCATED)[0]
        for amphora in amphorae:
            self.task_utils.mark_amphora_status_error(amphora.id)


class AmphoraPostVIPPlug(BaseAmphoraTask):
    """Task to notify the amphora post VIP plug."""

    def execute(self, amphora, loadbalancer, amphorae_network_config):
        """Execute post_vip_routine."""
        self.amphora_driver.post_vip_plug(
            amphora, loadbalancer, amphorae_network_config)
        LOG.debug("Notified amphora of vip plug")

    def revert(self, result, amphora, loadbalancer, *args, **kwargs):
        """Handle a failed amphora vip plug notification."""
        if isinstance(result, failure.Failure):
            return
        LOG.warning("Reverting post vip plug.")
        self.task_utils.mark_amphora_status_error(amphora.id)


class AmphoraePostVIPPlug(BaseAmphoraTask):
    """Task to notify the amphorae post VIP plug."""

    def execute(self, loadbalancer, amphorae_network_config):
        """Execute post_vip_plug across the amphorae."""
        amp_post_vip_plug = AmphoraPostVIPPlug()
        for amphora in loadbalancer.amphorae:
            amp_post_vip_plug.execute(amphora,
                                      loadbalancer,
                                      amphorae_network_config)


class AmphoraCertUpload(BaseAmphoraTask):
    """Upload a certificate to the amphora."""

    def execute(self, amphora, server_pem):
        """Execute cert_update_amphora routine."""
        LOG.debug("Upload cert in amphora REST driver")
        key = utils.get_compatible_server_certs_key_passphrase()
        fer = fernet.Fernet(key)
        self.amphora_driver.upload_cert_amp(amphora, fer.decrypt(server_pem))


class AmphoraUpdateVRRPInterface(BaseAmphoraTask):
    """Task to get and update the VRRP interface device name from amphora."""

    def execute(self, amphora, timeout_dict=None):
        try:
            interface = self.amphora_driver.get_interface_from_ip(
                amphora, amphora.vrrp_ip, timeout_dict=timeout_dict)
        except Exception as e:
            # This can occur when an active/standby LB has no listener
            LOG.error('Failed to get amphora VRRP interface on amphora '
                      '%s. Skipping this amphora as it is failing due to: '
                      '%s', amphora.id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora.id,
                                     status=constants.ERROR)
            return None

        self.amphora_repo.update(db_apis.get_session(), amphora.id,
                                 vrrp_interface=interface)
        return interface


class AmphoraIndexUpdateVRRPInterface(BaseAmphoraTask):
    """Task to get and update the VRRP interface device name from amphora."""

    def execute(self, amphora_index, amphorae, timeout_dict=None):
        amphora_id = amphorae[amphora_index].id
        try:
            interface = self.amphora_driver.get_interface_from_ip(
                amphorae[amphora_index], amphorae[amphora_index].vrrp_ip,
                timeout_dict=timeout_dict)
        except Exception as e:
            # This can occur when an active/standby LB has no listener
            LOG.error('Failed to get amphora VRRP interface on amphora '
                      '%s. Skipping this amphora as it is failing due to: '
                      '%s', amphora_id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                     status=constants.ERROR)
            return None

        self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                 vrrp_interface=interface)
        return interface


class AmphoraVRRPUpdate(BaseAmphoraTask):
    """Task to update the VRRP configuration of an amphora."""

    def execute(self, loadbalancer_id, amphorae_network_config, amphora,
                amp_vrrp_int, timeout_dict=None):
        """Execute update_vrrp_conf."""
        loadbalancer = self.loadbalancer_repo.get(db_apis.get_session(),
                                                  id=loadbalancer_id)
        # Note, we don't want this to cause a revert as it may be used
        # in a failover flow with both amps failing. Skip it and let
        # health manager fix it.
        amphora.vrrp_interface = amp_vrrp_int
        try:
            self.amphora_driver.update_vrrp_conf(
                loadbalancer, amphorae_network_config, amphora, timeout_dict)
        except Exception as e:
            LOG.error('Failed to update VRRP configuration amphora %s. '
                      'Skipping this amphora as it is failing to update due '
                      'to: %s', amphora.id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora.id,
                                     status=constants.ERROR)

        LOG.debug("Uploaded VRRP configuration of amphora %s.", amphora.id)


class AmphoraIndexVRRPUpdate(BaseAmphoraTask):
    """Task to update the VRRP configuration of an amphora."""

    def execute(self, loadbalancer_id, amphorae_network_config, amphora_index,
                amphorae, amp_vrrp_int, timeout_dict=None):
        """Execute update_vrrp_conf."""
        loadbalancer = self.loadbalancer_repo.get(db_apis.get_session(),
                                                  id=loadbalancer_id)
        # Note, we don't want this to cause a revert as it may be used
        # in a failover flow with both amps failing. Skip it and let
        # health manager fix it.
        amphora_id = amphorae[amphora_index].id
        amphorae[amphora_index].vrrp_interface = amp_vrrp_int
        try:
            self.amphora_driver.update_vrrp_conf(
                loadbalancer, amphorae_network_config, amphorae[amphora_index],
                timeout_dict)
        except Exception as e:
            LOG.error('Failed to update VRRP configuration amphora %s. '
                      'Skipping this amphora as it is failing to update due '
                      'to: %s', amphora_id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                     status=constants.ERROR)
            return
        LOG.debug("Uploaded VRRP configuration of amphora %s.", amphora_id)


class AmphoraVRRPStart(BaseAmphoraTask):
    """Task to start keepalived on an amphora.

    This will reload keepalived if it is already running.
    """

    def execute(self, amphora, timeout_dict=None):
        self.amphora_driver.start_vrrp_service(amphora, timeout_dict)
        LOG.debug("Started VRRP on amphora %s.", amphora.id)


class AmphoraIndexVRRPStart(BaseAmphoraTask):
    """Task to start keepalived on an amphora.

    This will reload keepalived if it is already running.
    """

    def execute(self, amphora_index, amphorae, timeout_dict=None):
        amphora_id = amphorae[amphora_index].id
        try:
            self.amphora_driver.start_vrrp_service(amphorae[amphora_index],
                                                   timeout_dict)
        except Exception as e:
            LOG.error('Failed to start VRRP on amphora %s. '
                      'Skipping this amphora as it is failing to start due '
                      'to: %s', amphora_id, str(e))
            self.amphora_repo.update(db_apis.get_session(), amphora_id,
                                     status=constants.ERROR)
            return
        LOG.debug("Started VRRP on amphora %s.", amphorae[amphora_index].id)


class AmphoraComputeConnectivityWait(BaseAmphoraTask):
    """Task to wait for the compute instance to be up."""

    def execute(self, amphora):
        """Execute get_info routine for an amphora until it responds."""
        try:
            amp_info = self.amphora_driver.get_info(amphora)
            LOG.debug('Successfuly connected to amphora %s: %s',
                      amphora.id, amp_info)
        except driver_except.TimeOutException:
            LOG.error("Amphora compute instance failed to become reachable. "
                      "This either means the compute driver failed to fully "
                      "boot the instance inside the timeout interval or the "
                      "instance is not reachable via the lb-mgmt-net.")
            self.amphora_repo.update(db_apis.get_session(), amphora.id,
                                     status=constants.ERROR)
            raise


class AmphoraConfigUpdate(BaseAmphoraTask):
    """Task to push a new amphora agent configuration to the amphora."""

    def execute(self, amphora, flavor):
        # Extract any flavor based settings
        if flavor:
            topology = flavor.get(constants.LOADBALANCER_TOPOLOGY,
                                  CONF.controller_worker.loadbalancer_topology)
        else:
            topology = CONF.controller_worker.loadbalancer_topology

        # Build the amphora agent config
        agent_cfg_tmpl = agent_jinja_cfg.AgentJinjaTemplater()
        agent_config = agent_cfg_tmpl.build_agent_config(amphora.id, topology)

        # Push the new configuration to the amphora
        try:
            self.amphora_driver.update_amphora_agent_config(amphora,
                                                            agent_config)
        except driver_except.AmpDriverNotImplementedError:
            LOG.error('Amphora %s does not support agent configuration '
                      'update. Please update the amphora image for this '
                      'amphora. Skipping.', amphora.id)



# Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

from octavia.common import constants
from octavia.controller.worker.v2 import controller_worker

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class Endpoints(object):

    # API version history:
    #   1.0 - Initial version.
    #   2.0 - Provider driver format.
    target = messaging.Target(
        namespace=constants.RPC_NAMESPACE_CONTROLLER_AGENT,
        version='2.0')

    def __init__(self):
        self.worker = controller_worker.ControllerWorker()

    def create_load_balancer(self, context, loadbalancer,
                             flavor=None, availability_zone=None):
        LOG.info('Creating load balancer \'%s\'...',
                 loadbalancer[constants.LOADBALANCER_ID])
        self.worker.create_load_balancer(loadbalancer, flavor,
                                         availability_zone)

    def update_load_balancer(self, context, original_load_balancer,
                             load_balancer_updates):
        LOG.info('Updating load balancer \'%s\'...',
                 original_load_balancer.get(constants.LOADBALANCER_ID))
        self.worker.update_load_balancer(original_load_balancer,
                                         load_balancer_updates)

    def delete_load_balancer(self, context, loadbalancer, cascade=False):
        LOG.info('Deleting load balancer \'%s\'...',
                 loadbalancer.get(constants.LOADBALANCER_ID))
        self.worker.delete_load_balancer(loadbalancer, cascade)

    def failover_load_balancer(self, context, load_balancer_id):
        LOG.info('Failing over amphora in load balancer \'%s\'...',
                 load_balancer_id)
        self.worker.failover_loadbalancer(load_balancer_id)

    def failover_amphora(self, context, amphora_id):
        LOG.info('Failing over amphora \'%s\'...',
                 amphora_id)
        self.worker.failover_amphora(amphora_id)

    def create_listener(self, context, listener):
        LOG.info('Creating listener \'%s\'...',
                 listener[constants.LISTENER_ID])
        self.worker.create_listener(listener)

    def update_listener(self, context, original_listener, listener_updates):
        LOG.info('Updating listener \'%s\'...',
                 original_listener[constants.LISTENER_ID])
        self.worker.update_listener(original_listener, listener_updates)

    def delete_listener(self, context, listener):
        LOG.info('Deleting listener \'%s\'...',
                 listener[constants.LISTENER_ID])
        self.worker.delete_listener(listener)

    def create_pool(self, context, pool):
        LOG.info('Creating pool \'%s\'...', pool.get(constants.POOL_ID))
        self.worker.create_pool(pool)

    def update_pool(self, context, original_pool, pool_updates):
        LOG.info('Updating pool \'%s\'...',
                 original_pool.get(constants.POOL_ID))
        self.worker.update_pool(original_pool, pool_updates)

    def delete_pool(self, context, pool):
        LOG.info('Deleting pool \'%s\'...', pool.get(constants.POOL_ID))
        self.worker.delete_pool(pool)

    def create_health_monitor(self, context, health_monitor):
        LOG.info('Creating health monitor \'%s\'...', health_monitor.get(
            constants.HEALTHMONITOR_ID))
        self.worker.create_health_monitor(health_monitor)

    def update_health_monitor(self, context, original_health_monitor,
                              health_monitor_updates):
        LOG.info('Updating health monitor \'%s\'...',
                 original_health_monitor.get(constants.HEALTHMONITOR_ID))
        self.worker.update_health_monitor(original_health_monitor,
                                          health_monitor_updates)

    def delete_health_monitor(self, context, health_monitor):
        LOG.info('Deleting health monitor \'%s\'...', health_monitor.get(
            constants.HEALTHMONITOR_ID))
        self.worker.delete_health_monitor(health_monitor)

    def create_member(self, context, member):
        LOG.info('Creating member \'%s\'...', member.get(constants.MEMBER_ID))
        self.worker.create_member(member)

    def update_member(self, context, original_member, member_updates):
        LOG.info('Updating member \'%s\'...', original_member.get(
            constants.MEMBER_ID))
        self.worker.update_member(original_member, member_updates)

    def batch_update_members(self, context, old_members, new_members,
                             updated_members):
        updated_member_ids = [m.get(constants.ID) for m in updated_members]
        new_member_ids = [m.get(constants.ID) for m in new_members]
        old_member_ids = [m.get(constants.ID) for m in old_members]
        LOG.info(
            'Batch updating members: old=\'%(old)s\', new=\'%(new)s\', '
            'updated=\'%(updated)s\'...',
            {'old': old_member_ids, 'new': new_member_ids,
             'updated': updated_member_ids})
        self.worker.batch_update_members(
            old_members, new_members, updated_members)

    def delete_member(self, context, member):
        LOG.info('Deleting member \'%s\'...', member.get(constants.MEMBER_ID))
        self.worker.delete_member(member)

    def create_l7policy(self, context, l7policy):
        LOG.info('Creating l7policy \'%s\'...',
                 l7policy.get(constants.L7POLICY_ID))
        self.worker.create_l7policy(l7policy)

    def update_l7policy(self, context, original_l7policy, l7policy_updates):
        LOG.info('Updating l7policy \'%s\'...', original_l7policy.get(
            constants.L7POLICY_ID))
        self.worker.update_l7policy(original_l7policy, l7policy_updates)

    def delete_l7policy(self, context, l7policy):
        LOG.info('Deleting l7policy \'%s\'...', l7policy.get(
            constants.L7POLICY_ID))
        self.worker.delete_l7policy(l7policy)

    def create_l7rule(self, context, l7rule):
        LOG.info('Creating l7rule \'%s\'...', l7rule.get(constants.L7RULE_ID))
        self.worker.create_l7rule(l7rule)

    def update_l7rule(self, context, original_l7rule, l7rule_updates):
        LOG.info('Updating l7rule \'%s\'...', original_l7rule.get(
            constants.L7RULE_ID))
        self.worker.update_l7rule(original_l7rule, l7rule_updates)

    def delete_l7rule(self, context, l7rule):
        LOG.info('Deleting l7rule \'%s\'...', l7rule.get(constants.L7RULE_ID))
        self.worker.delete_l7rule(l7rule)

    def update_amphora_agent_config(self, context, amphora_id):
        LOG.info('Updating amphora \'%s\' agent configuration...',
                 amphora_id)
        self.worker.update_amphora_agent_config(amphora_id)

    def delete_amphora(self, context, amphora_id):
        LOG.info('Deleting amphora \'%s\'...', amphora_id)
        self.worker.delete_amphora(amphora_id)



# Copyright 2018 Rackspace US Inc.  All rights reserved.
# Copyright 2019 Verizon Media
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia.common import constants as consts

# This is a JSON schema validation dictionary
# https://json-schema.org/latest/json-schema-validation.html
#
# Note: This is used to generate the amphora driver "supported availability
#       zone metadata" dictionary. Each property should include a description
#       for the user to understand what this availability zone setting does.
#
# Where possible, the property name should match the configuration file name
# for the setting. The configuration file setting is the default when a
# setting is not defined in an availability zone profile.

SUPPORTED_AVAILABILITY_ZONE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Octavia Amphora Driver Availability Zone Metadata Schema",
    "description": "This schema is used to validate new availability zone "
                   "profiles submitted for use in an amphora driver "
                   "availability zone.",
    "type": "object",
    "additionalProperties": False,
    "properties": {
        consts.COMPUTE_ZONE: {
            "type": "string",
            "description": "The compute availability zone."
        },
        consts.MANAGEMENT_NETWORK: {
            "type": "string",
            "description": "The management network ID for the amphora."
        },
        consts.VALID_VIP_NETWORKS: {
            "type": "array",
            "description": "List of network IDs that are allowed for VIP use. "
                           "This overrides/replaces the list of allowed "
                           "networks configured in `octavia.conf`."
        }
    }
}



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_config import cfg

from octavia.amphorae.backends.utils import interface
from octavia.common import config
from octavia.common import exceptions


class InterfaceException(exceptions.OctaviaException):
    message = "Could not configure interface: %(msg)s"


def interfaces_find(interface_controller, name):
    all_interfaces = interface_controller.list()

    if name == "all":
        return all_interfaces.values()

    if name in all_interfaces:
        return [all_interfaces[name]]

    msg = "Could not find interface '{}'.".format(name)
    raise InterfaceException(msg=msg)


def interfaces_update(interfaces, action_fn, action_str):
    errors = []

    for iface in interfaces:
        try:
            action_fn(iface)
        except Exception as e:
            errors.append("Error on action '{}' for interface {}: {}.".format(
                action_str, iface.name, e))

    if errors:
        raise InterfaceException(msg=", ".join(errors))


def interface_cmd(interface_name, action):
    interface_controller = interface.InterfaceController()

    if action == "up":
        action_fn = interface_controller.up
    elif action == "down":
        action_fn = interface_controller.down
    else:
        raise InterfaceException(
            msg="Unknown action '{}'".format(action))

    interfaces = interfaces_find(interface_controller,
                                 interface_name)
    interfaces_update(interfaces, action_fn, action)


def main():
    config.init(sys.argv[1:-2])
    config.setup_logging(cfg.CONF)

    try:
        action = sys.argv[-2]
        interface_name = sys.argv[-1]
    except IndexError:
        print("usage: {} [up|down] <interface>".format(sys.argv[0]))
        sys.exit(1)

    try:
        interface_cmd(interface_name, action)
    except Exception as e:
        print("Error: {}".format(e))
        sys.exit(2)


if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob

import docutils.core
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
import testtools


class FakeDirective(rst.Directive):
    has_content = True

    def run(self):
        return []


def fake_role(name, rawtext, text, lineno, inliner,
              options=None, content=None):
    return [], []


directives.register_directive('seqdiag', FakeDirective)
directives.register_directive('blockdiag', FakeDirective)
directives.register_directive('nwdiag', FakeDirective)
directives.register_directive('actdiag', FakeDirective)
directives.register_directive('graphviz', FakeDirective)
roles.register_local_role('doc', fake_role)


class TestTitles(testtools.TestCase):
    def _get_title(self, section_tree):
        section = {
            'subtitles': [],
        }
        for node in section_tree:
            if node.tagname == 'title':
                section['name'] = node.rawsource
            elif node.tagname == 'section':
                subsection = self._get_title(node)
                section['subtitles'].append(subsection['name'])
        return section

    def _get_titles(self, spec):
        titles = {}
        for node in spec:
            if node.tagname == 'section':
                section = self._get_title(node)
                titles[section['name']] = section['subtitles']
        return titles

    def _check_titles(self, titles):
        self.assertEqual(7, len(titles))

        problem = 'Problem description'
        self.assertIn(problem, titles)
        self.assertEqual(0, len(titles[problem]))

        proposed = 'Proposed change'
        self.assertIn(proposed, titles)
        self.assertIn('Alternatives', titles[proposed])
        self.assertIn('Data model impact', titles[proposed])
        self.assertIn('REST API impact', titles[proposed])
        self.assertIn('Security impact', titles[proposed])
        self.assertIn('Notifications impact', titles[proposed])
        self.assertIn('Other end user impact', titles[proposed])
        self.assertIn('Performance Impact', titles[proposed])
        self.assertIn('Other deployer impact', titles[proposed])
        self.assertIn('Developer impact', titles[proposed])

        impl = 'Implementation'
        self.assertIn(impl, titles)
        self.assertEqual(2, len(titles[impl]))
        self.assertIn('Assignee(s)', titles[impl])
        self.assertIn('Work Items', titles[impl])

        deps = 'Dependencies'
        self.assertIn(deps, titles)
        self.assertEqual(0, len(titles[deps]))

        testing = 'Testing'
        self.assertIn(testing, titles)
        self.assertEqual(0, len(titles[testing]))

        docs = 'Documentation Impact'
        self.assertIn(docs, titles)
        self.assertEqual(0, len(titles[docs]))

        refs = 'References'
        self.assertIn(refs, titles)
        self.assertEqual(0, len(titles[refs]))

    def test_template(self):
        files = set(glob.glob('specs/*.rst') + glob.glob('specs/*/*'))
        files = files - set(glob.glob('specs/*/*.dot'))
        files = files - set(glob.glob('specs/*/*.diag'))
        for filename in files:
            self.assertTrue(filename.endswith(".rst"),
                            "spec's file must use 'rst' extension.")
            with open(filename) as f:
                data = f.read()
            spec = docutils.core.publish_doctree(data)
            titles = self._get_titles(spec)
            self._check_titles(titles)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ipaddress
import os
import stat

from oslo_config import cfg
import simplejson

from octavia.common import constants as consts

CONF = cfg.CONF


class InterfaceFile(object):
    def __init__(self, name, mtu=None, addresses=None,
                 routes=None, rules=None, scripts=None):
        self.name = name
        self.mtu = mtu
        self.addresses = addresses or []
        self.routes = routes or []
        self.rules = rules or []
        self.scripts = scripts or {
            consts.IFACE_UP: [],
            consts.IFACE_DOWN: []
        }

    @classmethod
    def get_extensions(cls):
        return [".json"]

    @classmethod
    def load(cls, fp):
        return simplejson.load(fp)

    @classmethod
    def dump(cls, obj):
        return simplejson.dumps(obj)

    @classmethod
    def from_file(cls, filename):
        with open(filename, encoding='utf-8') as fp:
            config = cls.load(fp)

        return InterfaceFile(**config)

    @classmethod
    def get_directory(cls):
        return (CONF.amphora_agent.agent_server_network_dir or
                consts.AMP_NET_DIR_TEMPLATE)

    @classmethod
    def get_host_routes(cls, routes, **kwargs):
        host_routes = []
        if routes:
            for hr in routes:
                route = {
                    consts.DST: hr['destination'],
                    consts.GATEWAY: hr['nexthop'],
                    consts.FLAGS: [consts.ONLINK]
                }
                route.update(kwargs)
                host_routes.append(route)
        return host_routes

    def write(self):
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC

        net_dir = self.get_directory()

        try:
            os.makedirs(net_dir)
        except OSError:
            pass

        interface_file = "{}.json".format(self.name)

        with os.fdopen(os.open(os.path.join(net_dir, interface_file),
                               flags, mode), 'w') as fp:
            interface = {
                consts.NAME: self.name,
                consts.ADDRESSES: self.addresses,
                consts.ROUTES: self.routes,
                consts.RULES: self.rules,
                consts.SCRIPTS: self.scripts
            }
            if self.mtu:
                interface[consts.MTU] = self.mtu
            fp.write(self.dump(interface))


class VIPInterfaceFile(InterfaceFile):
    def __init__(self, name, mtu,
                 vip, ip_version, prefixlen,
                 gateway, vrrp_ip, host_routes,
                 topology):

        super().__init__(name, mtu=mtu)

        if vrrp_ip:
            self.addresses.append({
                consts.ADDRESS: vrrp_ip,
                consts.PREFIXLEN: prefixlen
            })
        else:
            key = consts.DHCP if ip_version == 4 else consts.IPV6AUTO
            self.addresses.append({
                key: True
            })

        if gateway:
            # Add default routes if there's a gateway
            self.routes.append({
                consts.DST: (
                    "::/0" if ip_version == 6 else "0.0.0.0/0"),
                consts.GATEWAY: gateway,
                consts.FLAGS: [consts.ONLINK]
            })
            self.routes.append({
                consts.DST: (
                    "::/0" if ip_version == 6 else "0.0.0.0/0"),
                consts.GATEWAY: gateway,
                consts.FLAGS: [consts.ONLINK],
                consts.TABLE: 1,
            })

        # In ACTIVE_STANDBY topology, keepalived sets these addresses, routes
        # and rules
        if topology == consts.TOPOLOGY_SINGLE:
            self.addresses.append({
                consts.ADDRESS: vip,
                consts.PREFIXLEN: prefixlen
            })
            vip_cidr = ipaddress.ip_network(
                "{}/{}".format(vip, prefixlen), strict=False)
            self.routes.append({
                consts.DST: vip_cidr.exploded,
                consts.PREFSRC: vip,
                consts.SCOPE: 'link',
                consts.TABLE: 1,
            })
            self.rules.append({
                consts.SRC: vip,
                consts.SRC_LEN: 128 if ip_version == 6 else 32,
                consts.TABLE: 1,
            })

        self.routes.extend(self.get_host_routes(host_routes))
        self.routes.extend(self.get_host_routes(host_routes,
                                                table=1))

        self.scripts[consts.IFACE_UP].append({
            consts.COMMAND: (
                "/usr/local/bin/lvs-masquerade.sh add {} {}".format(
                    'ipv6' if ip_version == 6 else 'ipv4', name))
        })
        self.scripts[consts.IFACE_DOWN].append({
            consts.COMMAND: (
                "/usr/local/bin/lvs-masquerade.sh delete {} {}".format(
                    'ipv6' if ip_version == 6 else 'ipv4', name))
        })


class PortInterfaceFile(InterfaceFile):
    def __init__(self, name, mtu, fixed_ips):
        super().__init__(name, mtu=mtu)

        if fixed_ips:
            ip_versions = set()

            for fixed_ip in fixed_ips:
                ip_addr = fixed_ip['ip_address']
                cidr = fixed_ip['subnet_cidr']
                ip = ipaddress.ip_address(ip_addr)
                network = ipaddress.ip_network(cidr)
                prefixlen = network.prefixlen
                self.addresses.append({
                    consts.ADDRESS: fixed_ip['ip_address'],
                    consts.PREFIXLEN: prefixlen,
                })

                ip_versions.add(ip.version)

                host_routes = self.get_host_routes(
                    fixed_ip.get('host_routes', []))
                self.routes.extend(host_routes)
        else:
            ip_versions = {4, 6}

            self.addresses.append({
                consts.DHCP: True,
                consts.IPV6AUTO: True
            })

        for ip_version in ip_versions:
            self.scripts[consts.IFACE_UP].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh add {} {}".format(
                        'ipv6' if ip_version == 6 else 'ipv4', name))
            })
            self.scripts[consts.IFACE_DOWN].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh delete {} {}".format(
                        'ipv6' if ip_version == 6 else 'ipv4', name))
            })



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""create healthmanager table

Revision ID: 92fe9857279
Revises: 256852d5ff7c
Create Date: 2015-01-22 16:58:23.440247

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '92fe9857279'
down_revision = '256852d5ff7c'


def upgrade():
    op.create_table(
        u'amphora_health',
        sa.Column(u'amphora_id', sa.String(36), nullable=False,
                  primary_key=True),
        sa.Column(u'last_update', sa.DateTime(timezone=True),
                  nullable=False)

    )



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from octavia.i18n import _

LOG = logging.getLogger(__name__)


def check_database_connection(session):
    """This is a simple database connection check function.

    It will do a simple no-op query (low overhead) against the sqlalchemy
    session passed in.

    :param session: A Sql Alchemy database session.
    :returns: True if the connection check is successful, False if not.
    """
    try:
        session.execute('SELECT 1;')
        return True, None
    except Exception as e:
        message = _('Database health check failed due to: {err}.').format(
            err=str(e))
        LOG.error(message)
        return False, message
